// Copyright (c) Huawei Technologies Co., Ltd. 2019. All rights reserved.
// iSulad-img licensed under the Mulan PSL v2.
// You can use this software according to the terms and conditions of the Mulan PSL v2.
// You may obtain a copy of Mulan PSL v2 at:
//     http://license.coscl.org.cn/MulanPSL2
// THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY OR FIT FOR A PARTICULAR
// PURPOSE.
// See the Mulan PSL v2 for more details.
// Description: AES Encrypt and Decrypt
// Author: wangfengtu
// Create: 2019-07-16

package aes

import (
	"crypto/aes"
	"crypto/cipher"
	"crypto/rand"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
)

var aesKey []byte

const (
	// Use AES-256
	keyLen           = 32
	defaultAESKeyDir = "/root/.isulad"
	aesKeyName       = "aeskey"
)

func genRandData(size int) ([]byte, error) {
	buf := make([]byte, size)
	_, err := rand.Read(buf)
	if err != nil {
		return nil, err
	}
	return buf, nil
}

// Init aes key, create key file if not exist
func Init(dir string) error {
	var filename string
	var key []byte

	if dir != "" {
		filename = filepath.Join(dir, aesKeyName)
	} else {
		filename = filepath.Join(defaultAESKeyDir, aesKeyName)
	}

	if _, err := os.Stat(filename); err == nil {
		if key, err = ioutil.ReadFile(filename); err != nil {
			return fmt.Errorf("Read AES key failed: %v", err)
		}
		if len(key) != keyLen {
			return fmt.Errorf("Invalid aes key length %v, it must be %v", len(key), keyLen)
		}
	} else if os.IsNotExist(err) {
		// Create key file if not exist
		key, err = genRandData(keyLen)
		if err != nil {
			return fmt.Errorf("Generate AES key failed: %v", err)
		}
		if err = ioutil.WriteFile(filename, key, 0600); err != nil {
			return fmt.Errorf("Write key to file failed: %v", err)
		}
	} else {
		return err
	}

	aesKey = key

	return nil
}

// Encrypt data using CFB mode to be compatiable with docker
func Encrypt(plainText []byte) ([]byte, error) {
	block, err := aes.NewCipher(aesKey)
	if err != nil {
		return nil, fmt.Errorf("Encrypt data failed: %v", err)
	}

	iv, err := genRandData(block.BlockSize())
	if err != nil {
		return nil, fmt.Errorf("Generate rand data for iv failed: %v", err)
	}
	encrypter := cipher.NewCFBEncrypter(block, iv)
	encryptData := make([]byte, len(plainText))
	encrypter.XORKeyStream(encryptData, plainText)

	return append(iv, encryptData...), nil
}

// Decrypt data
func Decrypt(secretText []byte) ([]byte, error) {
	block, err := aes.NewCipher(aesKey)
	if err != nil {
		return nil, fmt.Errorf("Decrypt data failed: %v", err)
	}

	if len(secretText) <= block.BlockSize() {
		return nil, fmt.Errorf("Invalid secretText length %v, it must be larger then %v",
			len(secretText), block.BlockSize)
	}

	iv := secretText[:block.BlockSize()]
	decrypter := cipher.NewCFBDecrypter(block, iv)
	decryptData := make([]byte, len(secretText)-block.BlockSize())
	decrypter.XORKeyStream(decryptData, secretText[block.BlockSize():])

	return decryptData, nil
}
